@@ -79,6 +79,16 @@ static bool php_phongo_manager_merge_context_options(zval* zdriverOptions TSRMLS
 		return false;
 	}
 
+	/* When running PHP in debug mode, php_error_docref duplicates the current
+	 * scope, leading to a COW violation in zend_hash_merge and
+	 * zend_symtable_str_del (called by php_array_unsetc). This macro allows
+	 * that violation in debug mode and is a NOOP when in non-debug. */
+#if PHP_VERSION_ID >= 70200
+	HT_ALLOW_COW_VIOLATION(Z_ARRVAL_P(zdriverOptions));
+#endif
+
+	php_error_docref(NULL TSRMLS_CC, E_DEPRECATED, "The \"context\" driver option is deprecated.");
+
 	/* Perform array union (see: add_function() in zend_operators.c) */
 #if PHP_VERSION_ID >= 70000
 	zend_hash_merge(Z_ARRVAL_P(zdriverOptions), Z_ARRVAL_P(zcontextOptions), zval_add_ref, 0);
@@ -90,6 +100,7 @@ static bool php_phongo_manager_merge_context_options(zval* zdriverOptions TSRMLS
 #endif
 
 	php_array_unsetc(zdriverOptions, "context");
+
 	return true;
 } /* }}} */
 
@@ -114,7 +125,7 @@ static void php_phongo_manager_prep_authmechanismproperties(zval* properties TSR
 		zend_ulong   num_key    = 0;
 		zval*        property;
 
-		ZEND_HASH_FOREACH_KEY_VAL(ht_data, num_key, string_key, property)
+		ZEND_HASH_FOREACH_KEY_VAL_IND(ht_data, num_key, string_key, property)
 		{
 			if (!string_key) {
 				continue;
@@ -193,7 +204,7 @@ static void php_phongo_manager_prep_uri_options(zval* options TSRMLS_DC) /* {{{
 		zend_ulong   num_key    = 0;
 		zval*        option;
 
-		ZEND_HASH_FOREACH_KEY_VAL(ht_data, num_key, string_key, option)
+		ZEND_HASH_FOREACH_KEY_VAL_IND(ht_data, num_key, string_key, option)
 		{
 			if (!string_key) {
 				continue;
@@ -338,6 +349,26 @@ static PHP_METHOD(Manager, __construct)
 	}
 } /* }}} */
 
+/* {{{ proto MongoDB\Driver\ClientEncryption MongoDB\Driver\Manager::createClientEncryption(array $options)
+   Return a ClientEncryption instance */
+static PHP_METHOD(Manager, createClientEncryption)
+{
+	php_phongo_manager_t*          intern;
+	php_phongo_clientencryption_t* clientencryption;
+	zval*                          options;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a", &options) == FAILURE) {
+		return;
+	}
+
+	intern = Z_MANAGER_OBJ_P(getThis());
+
+	object_init_ex(return_value, php_phongo_clientencryption_ce);
+	clientencryption = Z_CLIENTENCRYPTION_OBJ_P(return_value);
+
+	phongo_clientencryption_init(clientencryption, intern->client, options TSRMLS_CC);
+} /* }}} */
+
 /* {{{ proto MongoDB\Driver\Cursor MongoDB\Driver\Manager::executeCommand(string $db, MongoDB\Driver\Command $command[, array $options = null])
    Execute a Command */
 static PHP_METHOD(Manager, executeCommand)
@@ -376,6 +407,11 @@ static PHP_METHOD(Manager, executeCommand)
 		goto cleanup;
 	}
 
+	/* If the Manager was created in a different process, reset the client so
+	 * that cursors created by this process can be differentiated and its
+	 * session pool is cleared. */
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+
 	phongo_execute_command(intern->client, PHONGO_COMMAND_RAW, db, command, options, server_id, return_value, return_value_used TSRMLS_CC);
 
 cleanup:
@@ -419,6 +455,11 @@ static PHP_METHOD(Manager, executeReadCommand)
 		return;
 	}
 
+	/* If the Manager was created in a different process, reset the client so
+	 * that cursors created by this process can be differentiated and its
+	 * session pool is cleared. */
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+
 	phongo_execute_command(intern->client, PHONGO_COMMAND_READ, db, command, options, server_id, return_value, return_value_used TSRMLS_CC);
 } /* }}} */
 
@@ -451,6 +492,11 @@ static PHP_METHOD(Manager, executeWriteCommand)
 		return;
 	}
 
+	/* If the Manager was created in a different process, reset the client so
+	 * that cursors created by this process can be differentiated and its
+	 * session pool is cleared. */
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+
 	phongo_execute_command(intern->client, PHONGO_COMMAND_WRITE, db, command, options, server_id, return_value, return_value_used TSRMLS_CC);
 } /* }}} */
 
@@ -483,6 +529,11 @@ static PHP_METHOD(Manager, executeReadWriteCommand)
 		return;
 	}
 
+	/* If the Manager was created in a different process, reset the client so
+	 * that cursors created by this process can be differentiated and its
+	 * session pool is cleared. */
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+
 	phongo_execute_command(intern->client, PHONGO_COMMAND_READ_WRITE, db, command, options, server_id, return_value, return_value_used TSRMLS_CC);
 } /* }}} */
 
@@ -524,6 +575,11 @@ static PHP_METHOD(Manager, executeQuery)
 		goto cleanup;
 	}
 
+	/* If the Manager was created in a different process, reset the client so
+	 * that cursors created by this process can be differentiated and its
+	 * session pool is cleared. */
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+
 	phongo_execute_query(intern->client, namespace, query, options, server_id, return_value, return_value_used TSRMLS_CC);
 
 cleanup:
@@ -566,6 +622,10 @@ static PHP_METHOD(Manager, executeBulkWrite)
 		goto cleanup;
 	}
 
+	/* If the Server was created in a different process, reset the client so
+	 * that its session pool is cleared. */
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+
 	phongo_execute_bulk_write(intern->client, namespace, bulk, options, server_id, return_value, return_value_used TSRMLS_CC);
 
 cleanup:
@@ -739,6 +799,11 @@ static PHP_METHOD(Manager, startSession)
 		}
 	}
 
+	/* If the Manager was created in a different process, reset the client so
+	 * that its session pool is cleared. This will ensure that we do not re-use
+	 * a server session (i.e. LSID) created by a parent process. */
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+
 	cs = mongoc_client_start_session(intern->client, cs_opts, &error);
 
 	if (cs) {
@@ -760,6 +825,10 @@ ZEND_BEGIN_ARG_INFO_EX(ai_Manager___construct, 0, 0, 0)
 	ZEND_ARG_ARRAY_INFO(0, driverOptions, 0)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_INFO_EX(ai_Manager_createClientEncryption, 0, 0, 1)
+	ZEND_ARG_ARRAY_INFO(0, options, 0)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO_EX(ai_Manager_executeCommand, 0, 0, 2)
 	ZEND_ARG_INFO(0, db)
 	ZEND_ARG_OBJ_INFO(0, command, MongoDB\\Driver\\Command, 0)
@@ -798,6 +867,7 @@ ZEND_END_ARG_INFO()
 static zend_function_entry php_phongo_manager_me[] = {
 	/* clang-format off */
 	PHP_ME(Manager, __construct, ai_Manager___construct, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(Manager, createClientEncryption, ai_Manager_createClientEncryption, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Manager, executeCommand, ai_Manager_executeCommand, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Manager, executeReadCommand, ai_Manager_executeRWCommand, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Manager, executeWriteCommand, ai_Manager_executeRWCommand, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
@@ -830,6 +900,10 @@ static void php_phongo_manager_free_object(phongo_free_object_arg* object TSRMLS
 		intern->client = NULL;
 	}
 
+	if (intern->client_hash) {
+		efree(intern->client_hash);
+	}
+
 #if PHP_VERSION_ID < 70000
 	efree(intern);
 #endif
@@ -844,6 +918,8 @@ static phongo_create_object_retval php_phongo_manager_create_object(zend_class_e
 	zend_object_std_init(&intern->std, class_type TSRMLS_CC);
 	object_properties_init(&intern->std, class_type);
 
+	PHONGO_SET_CREATED_BY_PID(intern);
+
 #if PHP_VERSION_ID >= 70000
 	intern->std.handlers = &php_phongo_handler_manager;
 
@@ -882,7 +958,13 @@ static HashTable* php_phongo_manager_get_debug_info(zval* object, int* is_temp T
 	for (i = 0; i < n; i++) {
 		zval obj;
 
-		php_phongo_server_to_zval(&obj, sds[i]);
+		if (!php_phongo_server_to_zval(&obj, sds[i])) {
+			/* Exception already thrown */
+			zval_ptr_dtor(&obj);
+			zval_ptr_dtor(&cluster);
+			goto done;
+		}
+
 		add_next_index_zval(&cluster, &obj);
 	}
 
@@ -895,13 +977,20 @@ static HashTable* php_phongo_manager_get_debug_info(zval* object, int* is_temp T
 		zval* obj = NULL;
 
 		MAKE_STD_ZVAL(obj);
-		php_phongo_server_to_zval(obj, sds[i]);
+		if (!php_phongo_server_to_zval(obj, sds[i])) {
+			/* Exception already thrown */
+			zval_ptr_dtor(&obj);
+			zval_ptr_dtor(&cluster);
+			goto done;
+		}
+
 		add_next_index_zval(cluster, obj);
 	}
 
 	ADD_ASSOC_ZVAL_EX(&retval, "cluster", cluster);
 #endif
 
+done:
 	mongoc_server_descriptions_destroy_all(sds, n);
 
 	return Z_ARRVAL(retval);
